!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: rckeyvaluedir.gs,v 1.7 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy:
!   RcKeyValueDirectory, RcKeyValueDictionary, AbstractDictionary, Collection, 
!   Object.
!
!=========================================================================

run
  RcKeyValueDictionary subclass: 'RcKeyValueDirectory'
    instVarNames: #()
    classVars: #()
    poolDictionaries: #()
    inDictionary: Globals
    constraints: #()
    instancesInvariant: false
    isModifiable: false
%

removeallmethods RcKeyValueDirectory
removeallclassmethods RcKeyValueDirectory

category: 'For Documentation Installation only'
classmethod: RcKeyValueDirectory
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'Like an RcKeyValueDictionary, an RcKeyValueDirectory is a collection that
 provides Dictionary-like behavior, but whose implementation is optimized by
 hashing the key/value pairs.  RcKeyValueDirectory implements many of the same
 protocols as RcKeyValueDictionary.

 Like RcKeyValueDictionary, RcKeyValueDirectory provides for concurrent
 handling of an individual instance by multiple sessions.  Any or all of those
 sessions can modify the single instance.  When that happens,
 RcKeyValueDirectory also reduces the transaction conflicts that can arise among
 those sessions when they attempt to commit the instance to GemStone.  An
 RcKeyValueDirectory differs from an RcKeyValueDictionary in that concurrent
 readers and writers of an entry with the same key will conflict.'.
doc documentClassWith: txt.

self description: doc.
%

! ------------------- Instance methods for RcKeyValueDirectory
category: 'Accessing'
method: RcKeyValueDirectory
_at: aKey logging: aBoolean otherwise: aValue

"Returns the value that corresponds to aKey.  If no such key/value pair exists,
 returns the result of evaluating the zero-argument block aBlock.  If aBoolean
 is true, log the read operation in the redo log (to guarantee repeatable
 reads)."

| bkt val index |

aKey == nil ifTrue: [ ^ self _error: #rtErrNilKey ].

bkt := super _rcAt: (self hashFunction: aKey).
index := bkt searchForKey: aKey.
index ~~ nil
  ifTrue: [ " found the key "
    val := bkt valueAt: index.
    aBoolean
      ifTrue: [ 
        " put an entry in the redo log "
        self _logAt: aKey found: val inCollisionBucket: bkt.

        " put the collision bucket in the write set "
        bkt _addToWriteSet.
      ].
    ^ val
  ]
  ifFalse: [ " key not found "
    ^ aValue
  ]
%

category: 'Private'
method: RcKeyValueDirectory
_logAt: aKey found: aValue inCollisionBucket: aCollisionBkt

"Creates a log entry for reading the given key and value.  The collision bucket
 is the object on which a conflict may occur."

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_repeatReadAt:value:;
    argArray: (Array with: aKey with: aValue).
System redoLog addLogEntry: logEntry forConflictObject: aCollisionBkt
%

category: 'Private'
method: RcKeyValueDirectory
_repeatReadAt: aKey value: aValue

"Ensures that the value found at aKey is aValue."

" only successful reads are logged "
^ aValue == (self _at: aKey logging: false otherwise: #_incompletePathTraversal)
%

category: 'Accessing'
method: RcKeyValueDirectory
at: aKey ifAbsent: aBlock

"Returns the value that corresponds to aKey.  If no such key/value pair exists,
 returns the result of evaluating the zero-argument block aBlock."

| val |
val := self _at: aKey logging: true otherwise: #_incompletePathTraversal.

val == #_incompletePathTraversal
  ifTrue: [ " key not found "
    aBlock _validateClass: BlockClosure.
    ^ aBlock value
  ]
  ifFalse: [ " key found "
    ^ val
  ]
%

category: 'Accessing'
method: RcKeyValueDirectory
at: aKey otherwise: aValue

"Returns the value that corresponds to aKey.  If no such key/value pair exists,
 returns the result of evaluating the zero-argument block aBlock."

^ self _at: aKey logging: true otherwise: aValue
%

