!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: identitybag.gs,v 1.9 2008-05-07 20:54:38 bretlb Exp $
!
! Superclass Hierarchy:
!   IdentityBag, UnorderedCollection, Collection, Object.
!
!=========================================================================

removeallmethods IdentityBag
removeallclassmethods IdentityBag

category: 'For Documentation Installation only'
classmethod: IdentityBag
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'An IdentityBag is an UnorderedCollection in which any distinct object can occur
 any number of times.  Adding the same (identical) object to an IdentityBag
 multiple times simply causes it to occur multiple times in the IdentityBag.

 Since an IdentityBag is an identity-based collection, different (non-identical)
 but equivalent (equal) objects are treated as distinct from each other.  In
 Bags, they are not distinct.  Adding multiple equivalent objects to an
 IdentityBag yields an IdentityBag with multiple objects as elements, each
 occurring once.

 You can create subclasses of IdentityBag to restrict the kind of elements it
 contains.  When creating a subclass of IdentityBag, you must specify a class as
 the aConstraint argument.  This class is called the element kind of the new
 subclass.  For each instance of the new subclass, the class of each element
 must be of the element kind.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'Methods in this category are obsolete and are provided only for compatibility
 with earlier releases of GemStone.  They will be removed in a future release.'.
doc documentClassCategory: #'Backward Compatibility' with: txt.

self description: doc.
%

!==========================================================================
! Below this point: only method definitions in fileout order
!==========================================================================

! ------------------- Class methods for IdentityBag
category: 'Backward Compatibility'
classmethod: IdentityBag
elementKind

"Obsolete in GemStone 3.2."

^ self varyingConstraint
%
category: 'Modifying Classes'
classmethod: IdentityBag
isModifiable

"Returns true if the receiver and its Array of named instance variables are
 modifiable."

^ ((self isInvariant) | (instVarNames isInvariant) ) not
%
category: 'Instance Creation'
classmethod: IdentityBag
new: anInteger

"(R) Returns an instance of the receiver with sufficient space to hold 
 anInteger number of elements.

 For IdentityBags, this method is the same as #new."

^ self new.
%
! ------------------- Instance methods for IdentityBag
category: 'Set Arithmetic'
method: IdentityBag
* aBagOrSet

"Intersection.  Returns a kind of IdentityBag containing only the elements that
 are present in both the receiver and the argument aBagOrSet.

 The class of the result is the lowest class in the special hierarchy 
 (described below) of which both and receiver and argument are some kind:

 IdentityBag --> IdentitySet --> Set
 IdentityBag --> Bag
 IdentityBag --> RcIdentityBag*

 * If either the receiver or argument are a kind of RcIdentityBag,
 the result is an IdentityBag.

 If the result is a kind of Set, then each element that occurs in both the
 receiver and aBagOrSet occurs exactly once in the result.  If the result is a
 IdentityBag that is not an IdentitySet, and if an element occurs m times in the
 receiver and n times in the argument aBagOrSet, then the result contains the
 lesser of m or n occurrences of that element."

^ (aBagOrSet _isRcIdentityBag _or: [ aBagOrSet _isIdentityBag not ])
  ifTrue: [ self _primIntersect: aBagOrSet _asIdentityBag ]
  ifFalse: [ self _primIntersect: aBagOrSet ]
%
category: 'Set Arithmetic'
method: IdentityBag
+ aBagOrSet

"Union.  Returns a kind of IdentityBag containing exactly the elements that are
 present in either the receiver or the argument aBagOrSet.

 The class of the result is the lowest class in the special hierarchy 
 (described below) of which both and receiver and argument are some kind:

 IdentityBag --> IdentitySet --> Set
 IdentityBag --> Bag
 IdentityBag --> RcIdentityBag*

 * If either the receiver or argument are a kind of RcIdentityBag,
 the result is an IdentityBag.

 If the result is a kind of IdentitySet, then each element that occurs in either
 the receiver or aBagOrSet occurs exactly once in the result.  If the result is
 an IdentityBag that is not an IdentitySet, and if an element occurs m times in
 the receiver and n times in the argument aBagOrSet, then the result contains
 m + n occurrences of that element."

^ (aBagOrSet _isRcIdentityBag _or: [ aBagOrSet _isIdentityBag not ])
  ifTrue: [ self _primUnion: aBagOrSet _asIdentityBag ]
  ifFalse: [ self _primUnion: aBagOrSet ]
%
category: 'Set Arithmetic'
method: IdentityBag
- aBagOrSet

"Difference.  Returns a kind of IdentityBag containing exactly those elements of
 the receiver that have a greater number of occurrences in the receiver than in
 the argument aBagOrSet.  If an element occurs m times in the receiver and n
 times in aBagOrSet (where m >= n), then the result will contain m - n
 occurrences of that element.

 The class of the result is the class of the receiver."

^ (aBagOrSet _isRcIdentityBag _or: [ aBagOrSet _isIdentityBag not ])
  ifTrue: [ self _primDifference: aBagOrSet _asIdentityBag ]
  ifFalse: [ self _primDifference: aBagOrSet ]
%
category: 'Comparing'
method: IdentityBag
= aBag

"Returns true if all of the following conditions are true:

 1.  The receiver and aBag are of the same class.
 2.  The two collections are of the same size.
 3.  They have the same element kind.
 4.  Their public named instance variables are identical.
 5.  The elements of the receiver and aBag are identical.
 6.  Each element occurs the same number of times in the receiver and in aBag.

 Returns false otherwise."

<primitive: 800 "with rcvrClass">
self _primitiveFailed: #= .
^false
%
category: 'Adding'
method: IdentityBag
add: newObject

"Adds newObject to the receiver.  Has no effect if newObject is nil."

<primitive: 208>
self _primitiveFailed: #add: .
self _uncontinuableError
%
category: 'Adding'
method: IdentityBag
add: anObject withOccurrences: anInteger

"Includes anObject as an element of the receiver anInteger number of times.
 Generates an error if anObject is not a kind of the receiver's element kind.
 Has no effect if anObject is nil."

1 to: anInteger do: [:anInt | self add: anObject ]
%
category: 'Adding'
method: IdentityBag
addAll: aCollection

"Adds all of the elements of aCollection to the receiver."

^ aCollection _isRcIdentityBag
  ifTrue: [ super addAll: aCollection ]
  ifFalse: [ self _addAll: aCollection ]
%
category: 'Accessing'
method: IdentityBag
at: anIndex

"Returns the element of the receiver that is currently located at position
 anIndex.

 The elements of an IdentityBag are inherently unordered, and can change 
 position (index) when the IdentityBag is altered.  Thus, after an IdentityBag
 is altered, a given element may reside at a different index than before, and a
 given index may house a different element.  You should not infer an ordering
 for an IdentityBag's elements when you access them by index.

 This method is useful primarily as a code optimizer for iterating over all the
 elements of an IdentityBag (using a loop that runs the index from 1 to 
 the size of the IdentityBag).  

 The IdentityBag must not change during the iteration.  But the iteration may
 run faster than it would if you use other alternatives such as the do:
 method."

<primitive: 64>
(anIndex _isInteger)
   ifTrue: [ ^ self _error: #objErrBadOffsetIncomplete args: #[anIndex] ]
ifFalse: [ ^ self _errorNonIntegerIndex: anIndex ].
self _primitiveFailed: #at: .
self _uncontinuableError
%
category: 'Updating'
method: IdentityBag
at: anIndex put: anObject

"Disallowed.  Generates an error, since the elements of an IdentityBag are not
 externally accessible through numeric indices."

^ self _errorNotIndexable
%
category: 'Searching'
method: IdentityBag
collect: aBlock

"(R) Evaluates aBlock with each of the receiver's elements as the argument.
 Collects the resulting values into a collection of class specified by 
 sending #speciesForCollect message to the receiver and returns the new 
 collection.  The argument aBlock must be a one-argument block."

| result mySize |

result:= self speciesForCollect new.
mySize := self size.
1 to: mySize do: [ :i | result add: (aBlock value:(self _at:i)) ].
^ result
%
category: 'Repository Conversion'
method: IdentityBag
conversionRebuild

"Rebuilds an NSC to fix up internal sort order of the elements.
 Returns the receiver."

| aClone numInstVars myClass myConstraint |

myClass := self _class.
myConstraint := myClass varyingConstraint.

"If constrained to hold only Pom oops, no need to rebuild."
((myConstraint == Object) _or: [myConstraint isSubclassOf: Magnitude])
  ifFalse: [ ^ self ].

aClone := myClass new.
aClone assignToSegment: self segment.

"First copy the named instance variables."
numInstVars:= myClass allInstVarNames size.
(self _class firstPublicInstVar) to: numInstVars do: [ :k | 
  aClone instVarAt: k put: (self instVarAt: k)
  ].

"Now copy the elements."
1 to: self size do:[:j | | anElement |
  anElement := self _at: j .
  aClone add: anElement .
  ].
aClone _primitiveBecome: self.
^aClone
%
category: 'Enumerating'
method: IdentityBag
do: aBlock

"Evaluates aBlock with each of the receiver's elements as the argument.
 The argument aBlock must be a one-argument block."

| theSize |
theSize := self size.
1 to: theSize do: [:i | aBlock value: (self _at: i) ].
^ self
%
category: 'Comparing'
method: IdentityBag
hash

"Returns an Integer hash code for the receiver."

| mySize |
(mySize := self size) == 0
  ifTrue: [ ^12794 ]
  ifFalse: [
    ^(((self _at: 1) identityHash bitShift: -1)
      bitXor: (self _at: mySize) identityHash)
        bitXor: mySize
  ]
%
category: 'Searching'
method: IdentityBag
includes: anObject

"Returns true if the argument anObject is an element of the receiver.
 Returns false otherwise.  (Compare with includesValue:, which is based
 on equality.)"

<primitive: 65>
self _primitiveFailed: #includes: .
self _uncontinuableError
%
category: 'Searching'
method: IdentityBag
includesIdentical: anObject

"Returns true if the argument anObject is an element of the receiver.
 Returns false otherwise.  (Compare with includesValue:, which is based
 on equality.)"

<primitive: 65>
self _primitiveFailed: #includes: .
self _uncontinuableError
%
category: 'Searching'
method: IdentityBag
includesValue: anObject

"Returns true if the receiver contains an object of the same value as the
 argument, anObject.  Returns false otherwise.  (Compare with includes:,
 which is based on identity.)"

1 to: self size do: [ :each |
  (anObject = (self _at: each)) ifTrue: [ ^true ]
].
^ false
%
category: 'Set Arithmetic'
method: IdentityBag
intersect: aBagOrSet into: anotherBagOrSet
"Same as the * method, except the result of the intersection
 is put into anotherBagOrSet"

<primitive: 533>
aBagOrSet _validateClass: IdentityBag.
anotherBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #intersect:into: .
self _uncontinuableError
%
category: 'Set Arithmetic'
method: IdentityBag
intersectAll: anArrayOfSetsOrBags into: aBagOrSet
"Starts by sorting by ascending size anArrayOfSetsOrBags.  Then intersect the first
 element with the receiver.  Take that result and intersect it with the second element.
 Continue until all intersections are completed and place the result set into aBagOrSet.
 Returns aBagOrSet. 

 If the anArrayOfSetsOrBags is empty, aBagOrSet will be a copy of the receiver.

 This method will fail if any of the following conditions is true:
 o  the array size is greater than 2031.
 o  any element of the array is not an IdentitySet or IdentityBag.
 o  any element of the array is empty.
 o  aBagOrSet is not an IdentityBag or IdentitySet"

<primitive: 532>
anArrayOfSetsOrBags _validateClass: Array.
aBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #intersectAll:into: .
self _uncontinuableError
%
category: 'Searching'
method: IdentityBag
occurrencesOf: anObject

"Returns the number of the receiver's elements that are identical (==) to
 anObject."

<primitive: 67>
self _primitiveFailed: #occurrencesOf: .
self _uncontinuableError
%
category: 'Removing'
method: IdentityBag
remove: anObject

"Removes anObject from the receiver and returns anObject.  If anObject is
 present several times in the receiver, only one occurrence is removed.
 Generates an error if anObject is not in the receiver."

"Note: In GemStone 4.1, this method returned the receiver if an object
 equivalent to oldObject was found in the receiver."

<primitive: 207>  "primitive fails if anObject not found"
^ self _errorNotFound: anObject
%
category: 'Removing'
method: IdentityBag
remove: anObject ifAbsent: exceptionBlock

"Removes from the receiver an object that is identical to anObject and
 returns anObject.  If several elements of the receiver are identical to
 oldObject, only one instance is removed.  If oldObject is not present in
 the receiver, evaluates anExceptionBlock and returns the result of the
 evaluation."

"Note: In GemStone 4.1, this method returned the receiver."

<primitive: 209>  "primitive fails if anObject not found"
^ exceptionBlock value
%
category: 'Removing'
method: IdentityBag
removeAll: aCollection

"Removes one occurrence of each element of aCollection from the
 receiver and returns the receiver.  Generates an error if any
 element of aCollection is not present in the receiver."

"Note: In GemStone 4.1, this method returned the receiver."

^ aCollection _isRcIdentityBag
  ifTrue: [ self _removeAll: aCollection _asIdentityBag errIfAbsent: true ]
  ifFalse: [ self _removeAll: aCollection errIfAbsent: true ]
%
category: 'Removing'
method: IdentityBag
removeAllPresent: aCollection

"Removes from the receiver one occurrence of each element of aCollection that is
 also an element of the receiver.  Differs from removeAll: in that, if some
 elements of aCollection are not present in the receiver, no error is generated.
 Returns aCollection."

"Note: In GemStone 4.1, this method returned the receiver."

^ aCollection _isRcIdentityBag
  ifTrue: [ self _removeAll: aCollection _asIdentityBag errIfAbsent: false ]
  ifFalse: [ self _removeAll: aCollection errIfAbsent: false ]
%
category: 'Removing'
method: IdentityBag
removeIdentical: anObject

"Removes anObject from the receiver and returns anObject.  If anObject is
 present several times in the receiver, only one occurrence is removed.
 Generates an error if anObject is not in the receiver."

^ self remove: anObject
%
category: 'Removing'
method: IdentityBag
removeIdentical: anObject ifAbsent: exceptionBlock

"Removes from the receiver an object that is identical to anObject and
 returns anObject.  If several elements of the receiver are identical to
 oldObject, only one instance is removed.  If oldObject is not present in
 the receiver, evaluates anExceptionBlock and returns the result of the
 evaluation."

^ self remove: anObject ifAbsent: exceptionBlock.
%
category: 'Removing'
method: IdentityBag
removeIfPresent: anObject

"Removes anObject from the receiver and returns anObject.  If anObject is
 present several times in the receiver, only one occurrence is removed.
 Returns nil if anObject is missing from the receiver."

"Note: In GemStone 4.1, this method returned the receiver."

<primitive: 207>  "primitive fails if anObject not found"
^nil
%
category: 'Set Arithmetic'
method: IdentityBag
sizeOfIntersectionWith: aBagOrSet
"Perform the intersection and return only the size of the result set.
 For large collections this method is faster than performing the 
 intersection."

<primitive: 531>
aBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #sizeOfIntersectionWith: .
self _uncontinuableError
%
category: 'Set Arithmetic'
method: IdentityBag
sizeOfIntersectionWithAll: anArrayOfBagsOrSets
"Starts by sorting by ascending size anArrayOfSetsOrBags.  Then intersect the first
 element with the receiver.  Take that result and intersect it with the second element.
 Continue to all intersections are completed. Return the size of the resultant set.
 If anArrayOfBagsOrSets is empty, return the size of the receiver.

 This method will fail if any of the following is true:
 o the array size is greater than 2031.
 o any element of the array is not an IdentitySet or IdentityBag.
 o any element of the array is empty."

<primitive: 534>
anArrayOfBagsOrSets _validateClass: Array.
self _primitiveFailed: #sizeOfIntersectionWithAll: .
self _uncontinuableError
%
category: 'Searching'
method: IdentityBag
speciesForCollect

"Returns a class, an instance of which should be used as the result of
 collect: or other projections applied to the receiver."

^ IdentityBag
%
category: 'Adding'
method: IdentityBag
_addAll: aCollection

"Adds all of the elements of aCollection to the receiver.  Occurrences 
 of nil in aCollection are not added to the receiver."

<primitive: 225>
(self == aCollection) ifTrue: [ ^ self addAll: (aCollection copy) ].
aCollection do: [:each | self add: each].
^ self
%
category: 'Accessing'
method: IdentityBag
_at: anInteger

"Private.  Functionally equivalent to 'at:'."

"Used to implement the 'do:' method, which cannot be implemented as a
 primitive itself, since blocks cannot be sent as arguments to primitives."

<primitive: 64>
(anInteger _isInteger)
ifTrue: [ ^ self _error: #objErrBadOffsetIncomplete args: #[anInteger] ]
ifFalse: [ ^ self _errorNonIntegerIndex: anInteger ].
self _primitiveFailed: #_at: .
self _uncontinuableError
%
category: 'Updating'
method: IdentityBag
_basicAdd: anObject

"Same as IdentityBag | add:.  Has no effect if anObject is nil."

<primitive: 208>

self _primitiveFailed: #_basicAdd: .
self _uncontinuableError
%
category: 'Searching'
method: IdentityBag
_indexOf: anElement

"Returns the numerical index into the receiver of the given element (or nil if
 the element cannot be found)."

"Uses _at: to do the search and matches on equality."

1 to: self size do: [ :each |
    (self _at: each) = anElement
        ifTrue: [ ^each ]
].
^nil
%
category: 'Testing'
method: IdentityBag
_isIdentityBag

"Return whether the receiver is an identity-based collection."

^ true
%
category: 'Searching'
method: IdentityBag
_offsetOf: anObject

"Returns the first SmallInteger offset such that
 (self _at: (self _offSetOf: anObject)) == anObject.

 If the argument is not in the receiver, returns 0.  The result of this method
 is only valid while no insertions or deletions are made to the IdentityBag."

"Implemented with sequential scan of the receiver."

<primitive: 62>
self _primitiveFailed: #_offsetOf: .
self _uncontinuableError
%
category: 'Set Arithmetic'
method: IdentityBag
_primDifference: aBagOrSet

""

<primitive: 223>
aBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #_primDifference: .
self _uncontinuableError
%
category: 'Set Arithmetic'
method: IdentityBag
_primIntersect: aBagOrSet

""

<primitive: 222>
aBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #_primIntersect: .
self _uncontinuableError
%
category: 'Updating'
method: IdentityBag
_primitiveAt: anIndex put: anObject

"Disallowed.  Generates an error, since the elements of an IdentityBag are not
 externally accessible through numeric indices."

^ self _errorNotIndexable
%
category: 'Set Arithmetic'
method: IdentityBag
_primUnion: aBagOrSet

""

<primitive: 221 >
aBagOrSet _validateClass: IdentityBag.
self _primitiveFailed: #_primUnion: .
self _uncontinuableError
%
category: 'Reduced Conflict Support'
method: IdentityBag
_rcDifference: aBag

"Returns the set arithmetic difference between the receiver and aBag.  Also
 places the root object and interior nodes that are accessed in the
 reduced-conflict read set for both the receiver and aBag."

<primitive: 224>
aBag _validateClass: IdentityBag.
self _primitiveFailed: #_rcDifference: .
self _uncontinuableError
%
category: 'Reduced Conflict Support'
method: IdentityBag
_rcIncludes: anObject

"Returns true if anObject is contained identically in the receiver; returns
 false otherwise.  Also places the root object and any interior nodes that are
 accessed in the reduced-conflict read set."

<primitive: 66>
self _primitiveFailed: #_rcIncludes: .
self _uncontinuableError
%
category: 'Reduced Conflict Support'
method: IdentityBag
_rcIncludesValue: anObject

"Returns true if anObject is contained in the receiver using equality
 comparison; returns false otherwise.  Also places the the root object of the
 receiver and any interior nodes that are accessed in the reduced-conflict read
 set."

1 to: self size do: [ :each |
  (anObject = (self _rcAt: each)) ifTrue: [ ^true ]
].
^ false
%
category: 'Reduced Conflict Support'
method: IdentityBag
_rcOccurrencesOf: anObject

"Returns the number of occurrences of anObject in the receiver.  Places the
 root of the receiver and any interior nodes that are accessed in the
 reduced-conflict read set."

<primitive: 68>
self _primitiveFailed: #_rcOccurrencesOf: .
self _uncontinuableError
%
category: 'Removing'
method: IdentityBag
_remove: anObject

"Removes anObject from the receiver and returns anObject.  If anObject is
 present several times in the receiver, only one occurrence is removed.
 Generates an error if anObject is not in the receiver."

"Needed by certain subclasses where the super remove: message picks up an
 undesired reimplementation."

<primitive: 207>  "primitive fails if anObject not found"
^ self _errorNotFound: anObject
%
category: 'Removing'
method: IdentityBag
_removeAll: aCollection errIfAbsent: aBool

"For each element in aCollection, removes the one element that is identical
 from the receiver.  Returns aCollection if successful."

"Note: In GemStone 4.1, this method returned the receiver."

<primitive: 226>

(aCollection isKindOf: Array)
  ifFalse: [ 
    self _removeAll: (aCollection asArray) errIfAbsent: aBool.
    ^ aCollection
    ].

(aCollection isKindOf: CharacterCollection) ifTrue:[
  aBool
    ifTrue:[
      1 to: aCollection size do:[:j| self remove: (aCollection _at: j) ].
    ]
    ifFalse:[
      1 to: aCollection size do:
        [:j| self remove: (aCollection _at: j) ifAbsent:[] ].
    ].
  ^ aCollection
  ].

aCollection _validateClass: Collection .
self _primitiveFailed: #_removeAll:errIfAbsent: .
self _uncontinuableError
%
category: 'Reduced Conflict Support'
method: IdentityBag
_selectiveAbort

"Performs an abort operation on the receiver. That is, if the object is
 committed, it removes any changes made by the current transaction and allows
 access to the committed state of the object.  This is not allowed if the object
 participates in an index."

_indexedPaths ~~ nil
  ifTrue: [ self _error: #rtErrSelectiveAbort ].
^ super _selectiveAbort
%
